/*
 * $Id: modular.cc,v 1.8 2005/05/08 17:47:25 amaia Exp
 * 
 */
#include "modular.h"
using std::vector;
using namespace std;

CProxy_Main mainProxy;
CProxy_simulacao simulProxy;
CkChareID mainhandle;
config gConfig;

int debug_verbosity;


/**
 *	Deprecated. not called anymore.
 *
 */
void Main::done()
{
	CkExit();
}

/**
 *	Ctor.
 */
Main::Main(CkArgMsg *m)
{
	timer1_.start();

	simul_done_count_ = 0;
	gConfig.init( m->argc,m->argv );
	gConfig.read();

	N_ 		= gConfig.get_N();
	M_ 		= gConfig.get_M();
	T_ 		= gConfig.get_T();
	Ttrans_	= gConfig.get_Ttrans();
	tau_ 	= gConfig.get_tau();
	U_		= gConfig.get_U();
	VEZ_	= gConfig.get_VEZ();

	DE_ 	= 1. / gConfig.get_S2();
	Edim_ 	= static_cast<int>( gConfig.get_Emax() / DE_ );

	//these vectors are fixed-size
	P_energy_histo_		.resize( Edim_ );
	P_K_energy_histo_	.resize( Edim_ );
	P_energy_			.resize( Edim_ );
	P_K_energy_			.resize( Edim_ );
	PP_energy_			.resize( Edim_ );
	real_index_			.resize( Edim_ );
	real_K_index_		.resize( Edim_ );
	E_index_			.resize( Edim_ );

	//fixed-size also
	EK_med_			.resize( T_ );
	VM_med_			.resize( T_ );
	Vi_med_			.resize( T_ );
	m_med_			.resize( T_ );
	V_				.resize( T_ );
	K_				.resize( T_ );

	debug_verbosity = gConfig.get_debug_level();

	debug_print( DEBUG_INFO3, "N: %lld ", N_);
	debug_print( DEBUG_INFO3, "M: %lld ", M_);
	debug_print( DEBUG_INFO3, "Ttrans: %lld ", Ttrans_);
	debug_print( DEBUG_INFO3, "T: %lld ", T_);
	debug_print( DEBUG_INFO3, "VEZ: %lld ", VEZ_);
	debug_print( DEBUG_INFO3, "\n");
	debug_print( DEBUG_INFO3, "tau: %Lf ", tau_);
	debug_print( DEBUG_INFO3, "U: %Lf ", U_);
	debug_print( DEBUG_INFO3, "S2: %Lf ", gConfig.get_S2());
	debug_print( DEBUG_INFO3, "mtarget: %Lf ", gConfig.get_mtarget());
	debug_print( DEBUG_INFO3, "\n");
	debug_print( DEBUG_INFO3, "Emax: %Lf ", gConfig.get_Emax());
	debug_print( DEBUG_INFO3, "sigma_gauss_q: %Lf ", 
											gConfig.get_sigma_gauss_q());
	debug_print( DEBUG_INFO3, "sigma_gauss_p: %Lf ",
											gConfig.get_sigma_gauss_p());
	debug_print( DEBUG_INFO3, "\n\n");

	mainProxy = thishandle;
	mainhandle = thishandle;

	//register quiescence callback
	CkStartQD( CkIndex_Main::quiescence(), &mainhandle);

	simulArray_ = CProxy_simulacao::ckNew( VEZ_ );
	simulProxy = simulArray_ ;

	simulArray_.start();
	
}

void Main::quiescence()
{
	timer1_.stop();
	debug_print( DEBUG_INFO3, "finished (%f secs).\n", timer1_.get());
	CkExit();
}


/**
 * Callback called when a 'simulation' object finishes.
 * The mainchare receives the historical variables,
 * computes occup_index, dynamical averages, histo,
 * and print.
 */
void Main::cbSimulDone(
	const vdouble EK,
	const vdouble VM,
	const vdouble Vi,
	const vdouble m,
	const vdouble v_V,
	const vdouble v_K)
{
	assert( EK.size() == T_ );
	assert( VM.size() == T_ );
	assert( Vi.size() == T_ );
	assert( m.size() == T_ );
	assert( v_V.size() == T_ );
	assert( v_K.size() == T_ );

	vint occup_index		( Edim_ ),
		occup_index_histo	( Edim_ ),
		occup_K_index		( Edim_ ),
		occup_K_index_histo	( Edim_ );

	timer t1;
	t1.start();

	//based on "make_occup_index()"
	for( u64 i = 0; i < T_; i++ ) {
	
		int U_cell = static_cast<int>(
			floor( (EK[i] + VM[i] + DE_/2 ) / DE_) );
			
		int K_cell = static_cast<int>(
			floor( (EK[i] + DE_/2 ) /DE_) );

		assert( U_cell < Edim_ );

		occup_index			[U_cell]++;
		occup_index_histo	[U_cell]++;
		occup_K_index		[K_cell]++;
		occup_K_index_histo	[K_cell]++;
	
	}

	for( int i = 0; i < Edim_; i++ ) {
	
		debug_print(DEBUG_EXTRA,
			"\toccup_i[%d]=%d, occup_i_h[%d]=%d, occup_k_i[%d]=%d"
			", occup_k_i_h[%d]=%d\n", 
			i, occup_index[i],
			i, occup_index_histo[i],
			i, occup_K_index[i],
			i, occup_K_index_histo[i]);
	}


	//based on "make_E_index()"
	for( int i = 0; i < Edim_; i++) {
		E_index_[i] = DE_ * i;
	}

	update_histo_E( occup_index, occup_index_histo);
	update_histo_K( occup_K_index, occup_K_index_histo);
	
	V_ = v_V;
	K_ = v_K;
	
	update_dynamical_averages( EK, VM, Vi, m);
	print_dynamical_averages();
	
	t1.stop();
	
	debug_print(DEBUG_INFO2,
		"main::simuldone() - updated stats (%f secs)\n",
		t1.get());
		
	simul_done_count_++;

	return;
}

//@TODO: refactor. duplicated with update_histo_K
void Main::update_histo_E( vint& occup_index, vint& occup_index_histo )
{
	
	ldouble stddev_tmp;
	ostringstream filename;
	
	ios_base::openmode mode = (simul_done_count_) ? 
		ios_base::app : 
		ios_base::out;
	
	filename << gConfig.get_out_dir() << "/o_E.dat";
	ofstream ofs1(filename.str().c_str(),mode);

	const int count = simul_done_count_ + 1;

	ofs1 << "run " << count << "/" << gConfig.get_VEZ();
	ofs1 << " T/Ttrans:" << gConfig.get_T() << "/" << gConfig.get_Ttrans();
	ofs1 << "\n";
	ofs1 << "#index #E #P_energy #stdv_P_energy";
	ofs1 << "#counts #counts_acc #P_energy_histo\n";
	
	for(int i = 0; i < Edim_; i++) {
	
		debug_print(DEBUG_EXTRA, 
			"\t\t occup_index[%d]=%d\n", i, occup_index[i]);
			
		if (occup_index[i] > 0) {

			const int ridx = ++real_index_[i];
			const ldouble mean_occup = occup_index[i]/(1. * T_ );

			P_energy_[i] = 
				((ridx-1)*P_energy_[i] + mean_occup)
				/ (1.*ridx);

			PP_energy_[i] = 
				((ridx-1)*PP_energy_[i] + mean_occup * mean_occup)
				/ (1.*ridx);
		}
		
		debug_print(DEBUG_EXTRA,"\t\t P_energy[%d]=%6.3Lf\n", i, 
			P_energy_[i]);
			
		debug_print(DEBUG_EXTRA,"\t\t PP_energy[%d]=%6.3Lf\n", i, 
			PP_energy_[i]);
			
		debug_print(DEBUG_EXTRA, "\t\t occup_index_histo[%d]=%d\n", i,
			occup_index_histo[i]);
		
		if (occup_index_histo[i] > 0) {
			P_energy_histo_[i] = occup_index_histo[i]/(1. * T_ * count);
		}

		//write out occupation numbers
		stddev_tmp = PP_energy_[i] - P_energy_[i] * P_energy_[i];
		if( stddev_tmp < 0) {
			stddev_tmp *= -1.;
		}

		if( P_energy_histo_[i]) {		
			ofs1 << fixed << setprecision(8);
			ofs1 << i << " ";
			ofs1 << E_index_[i] << " ";
			ofs1 << P_energy_[i] << " ";
			ofs1 << sqrt(stddev_tmp)/2. << " ";
			ofs1 << occup_index[i]/(1.*T_) << " ";
			ofs1 << real_index_[i] << " ";
			ofs1 << P_energy_histo_[i] << "\n";
		}		
	}

	ofs1.close();
}

void Main::update_histo_K( vint& occup_index, vint& occup_index_histo)
{
	ldouble stddev_tmp;
	
	ostringstream filename;
	
	ios_base::openmode mode = 
		(simul_done_count_) ? 
		ios_base::app : 
		ios_base::out;
		
	filename << gConfig.get_out_dir() << "/o_K.dat";
	ofstream ofs1(filename.str().c_str(),mode);

	const int count = simul_done_count_ + 1;
	ofs1 << "run " << count << "/" << gConfig.get_VEZ(); 
	ofs1 << " T/Ttrans:" << gConfig.get_T() << "/" << gConfig.get_Ttrans() ;
	ofs1 << "\n";
	ofs1 << "#index #E #P_K_energy #counts #counts_acc #P_energy_histo\n";
	
	for( u64 i = 0; i < Edim_; i++) {
		if( occup_index[i] > 0) {
			const int ridx = ++real_K_index_[i];
			P_K_energy_[i] = 
				((ridx-1)*P_K_energy_[i]+(occup_index[i]/(1.*T_)))/(1.*ridx);
		}

		if( occup_index_histo[i] > 0) {
			P_K_energy_histo_[i] = occup_index_histo[i]/(1. * T_ *count);
		}

		//write out occupation numbers

		if( P_K_energy_histo_[i]) {
			ofs1 << fixed << setprecision(8);			
			ofs1 << i << " ";
			ofs1 << E_index_[i] << " ";
			ofs1 << P_K_energy_[i] << " ";
			ofs1 << occup_index[i]/(1.*T_) << " ";
			ofs1 << real_K_index_[i] << " ";
			ofs1 << P_K_energy_histo_[i] << "\n";
		}		
	}
	ofs1.close();
}

/**
 * Recebe como parametros os valores de EK,VM,Vi,m de uma
 * simulacao (todos timesteps) e atualiza as médias
 * EK_med, VM_med, Vi_med, m_med.
 */
void Main::update_dynamical_averages(
	const vdouble& EK, const vdouble& VM,
	const vdouble& Vi, const vdouble& m)
{
//	debug_print(DEBUG_INFO2,"\tMain::update_dynamical_averages\n"); 
	int vez = simul_done_count_ + 1;
	for( u64 i = 0; i < T_; i++) {
		EK_med_[i] = ((vez-1)* EK_med_[i]+EK[i])/(1.*vez);
		VM_med_[i] = ((vez-1)* VM_med_[i]+VM[i])/(1.*vez);
		Vi_med_[i] = ((vez-1)* Vi_med_[i]+Vi[i])/(1.*vez);
		 m_med_[i] = ((vez-1)*  m_med_[i]+ m[i])/(1.*vez);
	}

}

/**
 *
 *
 */
void Main::print_dynamical_averages()
{
	//debug_print(DEBUG_INFO2,"\tMain::print_dynamical_averages\n"); 
	//KM+ERR-Uxx-Nxx-Mxx

	ostringstream filename;
	
	ios_base::openmode mode = (simul_done_count_) ? 
		ios_base::app :
		ios_base::out;
		
	filename << gConfig.get_out_dir() << "/km+err.dat";
	ofstream ofs(filename.str().c_str(),mode);

	assert(N_ == gConfig.get_N());
	assert(U_ == gConfig.get_U());

	ofs << "vez=" << simul_done_count_ << "\n";
	for( u64 i = 0; i < T_; i++) {	
		ofs << fixed << setprecision(15);
		ofs << i << " ";
		ofs << 2. * EK_med_[i] / M_ << " ";
		ofs << (EK_med_[i]+VM_med_[i])/M_ << " ";
		ofs << m_med_[i] << " ";
		ofs << Vi_med_[i] << " ";
		ofs <<  100*fabsl(U_ - (K_[i]+V_[i])/(1.*N_))/U_ << " ";
		ofs << "\n";
	}
	ofs.close();
}


#include "modular.def.h"
