/*
 * $Id: modular.h,v 1.9 2005/08/22 01:04:35 amaia Exp $
 */
#ifndef __MODULAR_H
#define __MODULAR_H

#include <unistd.h>
#include <sys/time.h>
#include <string>
#include <vector>
#include <iomanip>
#include <cstdarg>
#include <cstdio>
#include <fstream>
#include <sstream>
#include "pup_stl.h"


typedef u_int64_t u64;
typedef long double ldouble;

typedef std::vector<ldouble> vdouble;
typedef std::vector<ldouble>::iterator vdouble_it;

typedef std::vector<int> vint;
typedef std::vector<int>::iterator vint_it;

#include "modular.decl.h"
#include "config.h"

enum {
	DEBUG_NONE=0,
	DEBUG_FATAL=1,
	DEBUG_WARN=2,
	DEBUG_INFO3=3,
	DEBUG_INFO2=4,
	DEBUG_INFO1=5,
	DEBUG_EXTRA=6
};


int load_pq(const std::string filename, vdouble& p, vdouble& q);
int save_pq(const std::string filename, vdouble& p, vdouble& q);
void debug_print_ts(int level);
void debug_print(int level, const char *fmt, ...);

CkReductionMsg *custom_sum_ldouble(int nMsg, CkReductionMsg **msgs);
CkReduction::reducerType sum_ldouble();

void register_sum_ldouble(void);

class timer {

private:
	struct timeval tv;
	struct timeval tv2;
	double delta;
	int running;

public:
		timer() :delta(0), running(true) { 
			gettimeofday(&tv, NULL); 
		}
		void start() {
			gettimeofday(&tv, NULL);
			running=true;
		}
		void stop() {
			gettimeofday(&tv2, NULL);
			running=false;
		}
		double get() {
			if(running==true) return -1;
			delta += (tv2.tv_sec - tv.tv_sec);
			delta += ((tv2.tv_usec - tv.tv_usec)/1.e6);
			return delta;
		}
};


class Main : public CBase_Main {

public:

	Main(CkArgMsg *m);
	void quiescence();
	void done();
	void cbSimulDone(
		const vdouble EK,
		const vdouble VM,
		const vdouble Vi,
		const vdouble m,
		const vdouble V,
		const vdouble K);

private:
	//local copies of config variables (for convenience)
	u64 N_;
	u64 M_;
	u64 T_;
	u64 Ttrans_;
	u64 VEZ_;
	ldouble tau_;
	ldouble U_;
	ldouble DE_;
	int Edim_;

	CProxy_simulacao simulArray_;
	int simul_done_count_;

	vdouble P_energy_, P_energy_histo_, 
		P_K_energy_, P_K_energy_histo_,
		PP_energy_,
		E_index_;

	vint real_index_,
		 real_K_index_;

	//hold mean values over time
	vdouble EK_med_,
		VM_med_,
		Vi_med_,
		m_med_,
	//@TODO: refactor. 
	//hold values from last simulation
	//should not be stored here!
		V_,
		K_;

	//to measure elapsed times
	timer timer1_;

private:
	void update_dynamical_averages(
		const vdouble& EK,
		const vdouble& VM,
		const vdouble& Vi,
		const vdouble& m);

	void print_dynamical_averages();
	//averages_and_print()
	void update_histo_E( vint& occup_index, vint& occup_index_histo);
	void update_histo_K( vint& occup_index, vint& occup_index_histo);

};

class rotor : public CBase_rotor
{

private:
	//CProxy_simulacao simulProxy;
	int simulIdx_;

	//consts, mas devido ao pup nao posso?
	u64  N_, M_;
	u64 T_, Ttrans_;
	ldouble tau_;

	//tamanho dos vetores p,q
	//qtdade de rotores por objeto
	int G_; 

	//numero efetivo de rotores para o obj 
	//(caso a N % G != 0, sera diferente
	//de G)
	int arrsize_;

	u64 t_;
	int step_;
	ldouble c_[4];
	ldouble d_[4];
	
	vdouble p_,q_;	
	ldouble K_, KM_;

	void init_consts();

	public:
	rotor(CkMigrateMessage *m);
	rotor(int simulIdx, int G);

	void set(vdouble p, vdouble q);

	void calc();
	void calc_stage1();
	void calc_stage2(CkReductionMsg* m);
	
	void pup(PUP::er &p);
	
	~rotor() {}

};

class simulacao : public CBase_simulacao {

public:
	simulacao(CkMigrateMessage* m);
	simulacao();
	void start();
	void cbDone1(CkReductionMsg* m);
	void cbDone2(CkReductionMsg* m);

	void pup(PUP::er &p);

private:
	void init_rotors(std::string input);

	/**
	 * \todo
	 * não é necessário manter esses atributos globais localmente.
	 * refactor.
	 *
	 * as classes com pup nao podem possuir atribs const?
	 */
	u64 N_, M_, Ttrans_, T_;
	ldouble tau_;

	/** quantidade de rotores por objeto Rotor */
	int G_;

	/** timestep atual */
	int t_;

	/** array de rotores que fazem parte da simulação */
	CProxy_rotor rotorArray_;

	/** temporizador */
	timer timer1_;

	/**
	 *
	 * historico dessas variaveis para todos os t
	 * dessa simulacao. sera usado posteriormente
	 * para calcular 'dynamical averages'(EK_med,VM_med,
	 * Vi_med, m_med).
	 *
	 * (as médias em Main são feitas para cada timestep,
	 *  em relacao à cada realização/simulação)
	 */
	vdouble v_EK_, v_VM_, v_Vi_, v_m_, v_V_, v_K_;
};


#endif
