/*
 * $Id: rotor.cc,v 1.4 2005/08/22 01:04:35 amaia Exp $
 */

#include "modular.h"

using namespace std;


extern CProxy_simulacao simulProxy;
extern config gConfig;
extern int debug_verbosity;

template<class T>
void PUParray(PUP::er& p, T* t, int nelem)
{
	for(int i=0; i<nelem; i++) p|t[i];
}


rotor::rotor(CkMigrateMessage *m)
{ 
	//Este construtor precisa ser criado
	//para dar suporte ao recurso de migracao.
	//a classe preicsa possuir rotina PUP tambem
}

rotor::rotor(int simulIdx, int G)
	:
	simulIdx_(simulIdx),
	G_(G),
	N_(gConfig.get_N()),
	M_(gConfig.get_M()),
	T_(gConfig.get_T()),
	Ttrans_(gConfig.get_Ttrans()),
	tau_(gConfig.get_tau()),
	p_(G_),
	q_(G_),
	step_(0)
{ 

	debug_print(DEBUG_EXTRA,
		"\t\trotor::rotor(simulIdx=%d, G=%d)\n", simulIdx, G);
		
	init_consts();

	//@TODO: refactor
	const int nelem = (int)(N_/G_) + (N_ % G_ == 0 ? 0 : 1);
	// o ultimo obj possui menos que G rotores
	arrsize_ = (thisIndex == nelem -1) ? N_ - thisIndex * G_ : G_;
	
}


/**
 * inicializacao das constantes de integracao
 */
void rotor::init_consts()
{
  /*Integration Constants*/
  c_[0] = 1.0/(ldouble)(2.0*(2.0-powl(2.0,1.0/3.0)));
  c_[1] = (1.0-powl(2.0,1.0/3.0))*c_[0];
  c_[2] = c_[1];
  c_[3] = c_[0];

  d_[0] = 1.0/(ldouble)(2.0-powl(2.0,1.0/3.0));
  d_[1] = -powl(2.0,1.0/3.0)*d_[0];
  d_[2] = d_[0];
  d_[3] = 0.;

#if 0
  debug_print(DEBUG_EXTRA,"Constantes de integracao:\n");
  debug_print(DEBUG_EXTRA,"c: %Lf %Lf %Lf %Lf\n",_c[0],_c[1],_c[2],_c[3]);
  debug_print(DEBUG_EXTRA,"d: %Lf %Lf %Lf %Lf\n",_d[0],_d[1],_d[2],_d[3]);
#endif

}

/**
 *
 */
void rotor::set(vdouble p, vdouble q)
{
	assert(p.size() == q.size());
	assert(p.size() == G_);
	assert(p.size() == arrsize_);

	p_.clear();
	q_.clear();
	
	p_.reserve(G_);
	q_.reserve(G_);

	copy( p.begin(), p.end(), back_inserter<vdouble>(p_));
	copy( q.begin(), q.end(), back_inserter<vdouble>(q_));

	assert(p_.size() == arrsize_);
	assert(q_.size() == arrsize_);
}

/**
 *
 */
void rotor::calc()
{	
	t_		= 0;
	step_	= 0;
	calc_stage1();
}



/**
 *	Estágio 1:
 *	- calcula q a partir do p
 *		- calcula as componentes de m devido a esse q
 *		- contribui esses valores de m para o reduction
 */
void rotor::calc_stage1()
{
	debug_print(DEBUG_INFO1,
		"stage1 - start - obj %d , t %d step %d\n", 
		thisIndex, t_, step_);
	
	ldouble mx=0, my=0, mx_M=0, my_M=0;

	//base indica o indice real do primeiro rotor deste objeto
	//@TODO: refactor
	const ldouble base = thisIndex * G_;

	assert( p_.size() == arrsize_);
	for (int i=0; i < arrsize_; i++) {
		
		q_[i]	+= c_[step_] * p_[i] * tau_;

		while (q_[i] < 0) {
			q_[i] += 2. * M_PI;
		}
		
		q_[i]	= fmod(q_[i] ,2. * M_PI);

		mx 		+= cosl(q_[i]);
		my 		+= sinl(q_[i]);
		mx_M 	+= (i+ base < M_) ? cosl(q_[i]) : 0;
		my_M 	+= (i+ base < M_) ? sinl(q_[i]) : 0;
	
	}

	mx   /= N_;
	my   /= N_;
	mx_M /= M_;
	my_M /= M_;

	CkCallback cb(CkIndex_rotor::calc_stage2(NULL), thisProxy);
	ldouble res[] = { mx, my, mx_M, my_M };
	contribute(sizeof(res), res, sum_ldouble(), cb);
	
	return;
}


/**
 *	Estágio 2:
 *		-recebe as somas das componentes m
 *		-calcula q a partir desses m (e F)
 */
void rotor::calc_stage2(CkReductionMsg* m)
{
	//@TODO: refactor
	const ldouble base = thisIndex * G_;
	
	ldouble mx,my,mx_M,my_M;
	ldouble *res = (ldouble*) m->getData();
	mx  	= *res++;
	my 		= *res++;
	mx_M	= *res++;
	my_M	= *res++;
	delete m;

	assert( p_.size() == arrsize_);

	K_ = .0;
	KM_ = .0;
	
	for( int i = 0; i < arrsize_; i++) {
	
		const ldouble F = -mx * sinl(q_[i]) + my * cosl(q_[i]);
		
		p_[i]	+= d_[step_] * F * tau_;
		
		if( step_ == 3) {
			//fourth step, calculation of p[i] is finished
			const ldouble k = .5 * p_[i] * p_[i];
			K_ += k;
			KM_ += ( i + base < M_ )? k : 0;
		}
	}

	ldouble V=.0 ,VM=.0 ,Vi=.0, m_temp=.0 ;
	
	if (step_ == 3 && thisIndex == 0) {

		V  = .5 * (1 - mx * mx - my * my) * N_;

		VM = .5 * M_ * (1 - mx_M * mx_M - my_M * my_M);
		
		Vi = M_ * (N_ - M_) 
			- M_ * ( mx_M * ( mx * N_ - mx_M * M_) 
			+ my_M * ( my * N_ - my_M * M_) );
			
		Vi /= M_ * ( N_ - M_ - 1);
		
		m_temp = mx_M * mx_M + my_M * my_M;
		
	}

	++step_;

	//cada iteracao possui 4 steps
	if (step_ < 4) {
	
//		debug_print("rotor::done_stage2() step=%d done\n",_step);
		calc_stage1();
		
	} else {

		//estou fazendo reduction com K, KM, apesar de esse valor
		//so importar no ultimo step (=3). melhorar isso..
		
		//debug_print(DEBUG_INFO1,
		//	"\t contrib 1 index=%d\n",thisIndex);

		CkCallback cb1(CkIndex_simulacao::cbDone1(NULL),
			CkArrayIndex1D(simulIdx_),
			simulProxy);
			
		ldouble res1[] = { V, VM, Vi, K_, KM_, m_temp};
		
		//debug_print(DEBUG_INFO1,
		//	"\t contribute(6 *sizeof(ldouble))=(6 * %d)=%d\n",
		//	sizeof(ldouble),sizeof(res1));
		contribute(sizeof(res1),res1,sum_ldouble(),cb1);
	
		//terminada essa iteração, o objeto continuara
		//fazendo t=t+1
		step_ = 0;
		t_++;
		if (t_ < T_) {
			calc_stage1();
		}
		//senao, fim da simulação
	}	


	return;
}


/**
 * PUP
 *
 */
void rotor::pup(PUP::er &p)
{
	ArrayElement1D::pup(p);
	p|G_;
	p|N_;
	p|M_;
	p|T_;
	p|Ttrans_;
	p|tau_;
	p|t_;
	p|step_;
	p|p_;
	p|q_;
	PUParray(p,c_,4);
	PUParray(p,d_,4);

}

