/*
 * $Id: util.cc,v 1.6 2005/08/21 15:20:19 amaia Exp $
 */
#include "modular.h"
#include <fstream>
#include <iomanip>

using namespace std;

extern int debug_verbosity;

/**
 * carrega os vetore p e q
 */
int load_pq(const std::string filename, vdouble& p, vdouble& q)
{
	p.clear();
	q.clear();
	
	ifstream ifs(filename.c_str());
	
	if (!ifs) {
		return -1;
	}

	ldouble tempa, tempb;
	
	ifs >> tempa >>  tempb;
	
	while (!ifs.eof()) {
		p.push_back(tempa);
		q.push_back(tempb);
		ifs >> tempa >> tempb;
	}
	return 0;
}

/**
 * salva os vetores p e q
 */
int save_pq(const std::string filename, vdouble& p, vdouble& q)
{
	assert(q.size() == p.size());
	
	ofstream ofs(filename.c_str());
	
	if (!ofs) {
		return -1;
	}

	ofs << fixed << setprecision(15);
	
	for (u64 i =0; i < p.size(); i++) {
		ofs << p[i] << " " << q[i] << "\n";
	}
	
	ofs.close();
	return 0;
}

/**
 *
 */
void debug_print(int level, const char *fmt, ...) {
#ifdef DEBUG
	if(debug_verbosity >= level) {
		char tmp[4096];
		va_list ap;
		va_start (ap, fmt);
		vsnprintf(tmp,1024,fmt,ap);
		va_end (ap);
		CkPrintf(tmp);
	}
#endif
}

/**
 *
 */
void debug_print_ts(int level)
{
	if(debug_verbosity >= level) {
		struct timeval tv;
		gettimeofday(&tv,NULL);
		CkPrintf("%ul %ul ",tv.tv_sec,tv.tv_usec);
	}
}

/**
 *
 */
CkReductionMsg *custom_sum_ldouble(int nMsg, CkReductionMsg **msgs)
{
	int numelem = msgs[0]->getSize()/sizeof(ldouble);
	
	ldouble result[numelem];
	
	//debug_print(DEBUG_INFO1,
	//	"custom_sum_ldouble nmsg=%d, numelem=%d\n", nMsg, numelem);
	
	for (int c=0; c < numelem; c++) {
		result[c] = 0;
	}

	for (int i=0; i < nMsg; i++) {
		ldouble *m = (ldouble*)msgs[i]->getData();
		for (int c=0; c < numelem; c++) {
			result[c]+= *m++;
		}
	}
	return CkReductionMsg::buildNew(numelem*sizeof(ldouble),result);
}

CkReduction::reducerType my_sum_ldouble;

/**
 *
 */
void register_sum_ldouble(void)
{
	my_sum_ldouble = CkReduction::addReducer(custom_sum_ldouble);
}

/**
 *
 */
CkReduction::reducerType sum_ldouble() 
{
	return my_sum_ldouble;
}

